@@ -3,8 +3,10 @@ use Mojo::Base 'Mojo::Server';
 
 use Carp 'croak';
 use Mojo::IOLoop;
+use Mojo::Transaction::WebSocket;
 use Mojo::URL;
 use Mojo::Util 'term_escape';
+use Mojo::WebSocket 'server_handshake';
 use Scalar::Util 'weaken';
 
 use constant DEBUG => $ENV{MOJO_DAEMON_DEBUG} || 0;
@@ -14,14 +16,17 @@ has [qw(backlog max_clients silent)];
 has inactivity_timeout => sub { $ENV{MOJO_INACTIVITY_TIMEOUT} // 15 };
 has ioloop => sub { Mojo::IOLoop->singleton };
 has listen => sub { [split ',', $ENV{MOJO_LISTEN} || 'http://*:3000'] };
-has max_requests => 25;
+has max_requests => 100;
 
 sub DESTROY {
   return if Mojo::Util::_global_destruction();
   my $self = shift;
-  $self->_remove($_) for keys %{$self->{connections} || {}};
   my $loop = $self->ioloop;
-  $loop->remove($_) for @{$self->acceptors};
+  $loop->remove($_) for keys %{$self->{connections} || {}}, @{$self->acceptors};
+}
+
+sub ports {
+  [map { $_[0]->ioloop->acceptor($_)->port } @{$_[0]->acceptors}];
 }
 
 sub run {
@@ -38,16 +43,17 @@ sub run {
 sub start {
   my $self = shift;
 
-  # Resume accepting connections
   my $loop = $self->ioloop;
   if (my $max = $self->max_clients) { $loop->max_connections($max) }
+
+  # Resume accepting connections
   if (my $servers = $self->{servers}) {
     push @{$self->acceptors}, $loop->acceptor(delete $servers->{$_})
       for keys %$servers;
   }
 
   # Start listening
-  else { $self->_listen($_) for @{$self->listen} }
+  elsif (!@{$self->acceptors}) { $self->_listen($_) for @{$self->listen} }
 
   return $self;
 }
@@ -72,24 +78,38 @@ sub _build_tx {
   my $tx = $self->build_tx->connection($id);
   $tx->res->headers->server('Mojolicious (Perl)');
   my $handle = $self->ioloop->stream($id)->handle;
-  $tx->local_address($handle->sockhost)->local_port($handle->sockport);
-  $tx->remote_address($handle->peerhost)->remote_port($handle->peerport);
+  unless ($handle->isa('IO::Socket::UNIX')) {
+    $tx->local_address($handle->sockhost)->local_port($handle->sockport);
+    $tx->remote_address($handle->peerhost)->remote_port($handle->peerport);
+  }
   $tx->req->url->base->scheme('https') if $c->{tls};
 
-  # Handle upgrades and requests
   weaken $self;
-  $tx->on(
-    upgrade => sub {
-      my ($tx, $ws) = @_;
-      $ws->server_handshake;
-      $self->{connections}{$id}{ws} = $ws;
-    }
-  );
   $tx->on(
     request => sub {
       my $tx = shift;
-      $self->emit(request => $self->{connections}{$id}{ws} || $tx);
+
+      my $req = $tx->req;
+      if (my $error = $req->error) { $self->_debug($id, $error->{message}) }
+
+      # WebSocket
+      if ($req->is_handshake) {
+        my $ws = $self->{connections}{$id}{next}
+          = Mojo::Transaction::WebSocket->new(handshake => $tx);
+        $self->emit(request => server_handshake $ws);
+      }
+
+      # HTTP
+      else { $self->emit(request => $tx) }
+
+      # Last keep-alive request or corrupted connection
+      my $c = $self->{connections}{$id};
+      $tx->res->headers->connection('close')
+        and ++Mojo::IOLoop->stream($id)->{closed}
+        if $c->{requests} >= $self->max_requests || $req->error;
+
       $tx->on(resume => sub { $self->_write($id) });
+      $self->_write($id);
     }
   );
 
@@ -99,13 +119,12 @@ sub _build_tx {
 
 sub _close {
   my ($self, $id) = @_;
-
-  # Finish gracefully
-  if (my $tx = $self->{connections}{$id}{tx}) { $tx->server_close }
-
+  if (my $tx = $self->{connections}{$id}{tx}) { $tx->closed }
   delete $self->{connections}{$id};
 }
 
+sub _debug { $_[0]->app->log->debug($_[2]) if $_[0]{connections}{$_[1]}{tx} }
+
 sub _finish {
   my ($self, $id) = @_;
 
@@ -115,31 +134,28 @@ sub _finish {
   return $self->_remove($id) if $tx->is_websocket;
 
   # Finish transaction
-  $tx->server_close;
+  delete($c->{tx})->closed;
 
   # Upgrade connection to WebSocket
-  if (my $ws = $c->{tx} = delete $c->{ws}) {
+  if (my $ws = delete $c->{next}) {
 
     # Successful upgrade
-    if ($ws->res->code == 101) {
+    if ($ws->handshake->res->code == 101) {
+      $c->{tx} = $ws->established(1);
       weaken $self;
       $ws->on(resume => sub { $self->_write($id) });
-      $ws->server_open;
+      $self->_write($id);
     }
 
     # Failed upgrade
-    else {
-      delete $c->{tx};
-      $ws->server_close;
-    }
+    else { $ws->closed }
   }
 
   # Close connection if necessary
-  my $req = $tx->req;
-  return $self->_remove($id) if $req->error || !$tx->keep_alive;
+  return $self->_remove($id) if $tx->error || !$tx->keep_alive;
 
   # Build new transaction for leftovers
-  return unless length(my $leftovers = $req->content->leftovers);
+  return unless length(my $leftovers = $tx->req->content->leftovers);
   $tx = $c->{tx} = $self->_build_tx($id, $c);
   $tx->server_read($leftovers);
 }
@@ -149,19 +165,24 @@ sub _listen {
 
   my $url   = Mojo::URL->new($listen);
   my $proto = $url->protocol;
-  croak qq{Invalid listen location "$listen"} unless $proto =~ /^https?$/;
+  croak qq{Invalid listen location "$listen"}
+    unless $proto eq 'http' || $proto eq 'https' || $proto eq 'http+unix';
 
-  my $query   = $url->query;
-  my $options = {
-    address => $url->host,
-    backlog => $self->backlog,
-    reuse   => $query->param('reuse')
-  };
-  if (my $port = $url->port) { $options->{port} = $port }
-  $options->{"tls_$_"} = $query->param($_) for qw(ca cert ciphers key);
+  my $query = $url->query;
+  my $options = {backlog => $self->backlog};
+  $options->{$_} = $query->param($_) for qw(fd single_accept reuse);
+  if ($proto eq 'http+unix') { $options->{path} = $url->host }
+  else {
+    if ((my $host = $url->host) ne '*') { $options->{address} = $host }
+    if (my $port = $url->port) { $options->{port} = $port }
+  }
+  $options->{"tls_$_"} = $query->param($_) for qw(ca ciphers version);
+  /^(.*)_(cert|key)$/ and $options->{"tls_$2"}{$1} = $query->param($_)
+    for @{$query->names};
+  if (my $cert = $query->param('cert')) { $options->{'tls_cert'}{''} = $cert }
+  if (my $key  = $query->param('key'))  { $options->{'tls_key'}{''}  = $key }
   my $verify = $query->param('verify');
   $options->{tls_verify} = hex $verify if defined $verify;
-  delete $options->{address} if $options->{address} eq '*';
   my $tls = $options->{tls} = $proto eq 'https';
 
   weaken $self;
@@ -169,7 +190,7 @@ sub _listen {
     $options => sub {
       my ($loop, $stream, $id) = @_;
 
-      my $c = $self->{connections}{$id} = {tls => $tls};
+      $self->{connections}{$id} = {tls => $tls};
       warn "-- Accept $id (@{[$stream->handle->peerhost]})\n" if DEBUG;
       $stream->timeout($self->inactivity_timeout);
 
@@ -177,8 +198,7 @@ sub _listen {
       $stream->on(error =>
           sub { $self && $self->app->log->error(pop) && $self->_close($id) });
       $stream->on(read => sub { $self->_read($id => pop) });
-      $stream->on(timeout =>
-          sub { $self->app->log->debug('Inactivity timeout') if $c->{tx} });
+      $stream->on(timeout => sub { $self->_debug($id, 'Inactivity timeout') });
     }
   );
 
@@ -186,25 +206,17 @@ sub _listen {
   $self->app->log->info(qq{Listening at "$url"});
   $query->pairs([]);
   $url->host('127.0.0.1') if $url->host eq '*';
-  say "Server available at $url";
+  say 'Server available at ', $options->{path} // $url;
 }
 
 sub _read {
   my ($self, $id, $chunk) = @_;
 
-  # Make sure we have a transaction and parse chunk
-  return unless my $c = $self->{connections}{$id};
+  # Make sure we have a transaction
+  my $c = $self->{connections}{$id};
   my $tx = $c->{tx} ||= $self->_build_tx($id, $c);
   warn term_escape "-- Server <<< Client (@{[_url($tx)]})\n$chunk\n" if DEBUG;
   $tx->server_read($chunk);
-
-  # Last keep-alive request or corrupted connection
-  $tx->res->headers->connection('close')
-    if (($c->{requests} || 0) >= $self->max_requests) || $tx->req->error;
-
-  # Finish or start writing
-  if    ($tx->is_finished) { $self->_finish($id) }
-  elsif ($tx->is_writing)  { $self->_write($id) }
 }
 
 sub _remove {
@@ -218,28 +230,16 @@ sub _url { shift->req->url->to_abs }
 sub _write {
   my ($self, $id) = @_;
 
-  # Get chunk and write
-  return unless my $c  = $self->{connections}{$id};
-  return unless my $tx = $c->{tx};
-  return if !$tx->is_writing || $c->{writing}++;
+  # Protect from resume event recursion
+  my $c = $self->{connections}{$id};
+  return if !(my $tx = $c->{tx}) || $c->{writing};
+  local $c->{writing} = 1;
   my $chunk = $tx->server_write;
-  delete $c->{writing};
   warn term_escape "-- Server >>> Client (@{[_url($tx)]})\n$chunk\n" if DEBUG;
-  my $stream = $self->ioloop->stream($id)->write($chunk);
-
-  # Finish or continue writing
+  my $next = $tx->is_finished ? '_finish' : length $chunk ? '_write' : undef;
+  return $self->ioloop->stream($id)->write($chunk) unless $next;
   weaken $self;
-  my $cb = sub { $self->_write($id) };
-  if ($tx->is_finished) {
-    if ($tx->has_subscribers('finish')) {
-      $cb = sub { $self->_finish($id) }
-    }
-    else {
-      $self->_finish($id);
-      return unless $c->{tx};
-    }
-  }
-  $stream->write('' => $cb);
+  $self->ioloop->stream($id)->write($chunk => sub { $self->$next($id) });
 }
 
 1;
@@ -255,8 +255,7 @@ Mojo::Server::Daemon - Non-blocking I/O HTTP and WebSocket server
   use Mojo::Server::Daemon;
 
   my $daemon = Mojo::Server::Daemon->new(listen => ['http://*:8080']);
-  $daemon->unsubscribe('request');
-  $daemon->on(request => sub {
+  $daemon->unsubscribe('request')->on(request => sub {
     my ($daemon, $tx) = @_;
 
     # Request
@@ -276,14 +275,14 @@ Mojo::Server::Daemon - Non-blocking I/O HTTP and WebSocket server
 =head1 DESCRIPTION
 
 L<Mojo::Server::Daemon> is a full featured, highly portable non-blocking I/O
-HTTP and WebSocket server, with IPv6, TLS, Comet (long polling), keep-alive and
-multiple event loop support.
+HTTP and WebSocket server, with IPv6, TLS, SNI, Comet (long polling), keep-alive
+and multiple event loop support.
 
 For better scalability (epoll, kqueue) and to provide non-blocking name
 resolution, SOCKS5 as well as TLS support, the optional modules L<EV> (4.0+),
 L<Net::DNS::Native> (0.15+), L<IO::Socket::Socks> (0.64+) and
 L<IO::Socket::SSL> (1.94+) will be used automatically if possible. Individual
-features can also be disabled with the C<MOJO_NO_NDN>, C<MOJO_NO_SOCKS> and
+features can also be disabled with the C<MOJO_NO_NNR>, C<MOJO_NO_SOCKS> and
 C<MOJO_NO_TLS> environment variables.
 
 See L<Mojolicious::Guides::Cookbook/"DEPLOYMENT"> for more.
@@ -309,9 +308,12 @@ implements the following new ones.
 =head2 acceptors
 
   my $acceptors = $daemon->acceptors;
-  $daemon       = $daemon->acceptors([]);
+  $daemon       = $daemon->acceptors(['6be0c140ef00a389c5d039536b56d139']);
 
-Active acceptors.
+Active acceptor ids.
+
+  # Check port
+  mu $port = $daemon->ioloop->acceptor($daemon->acceptors->[0])->port;
 
 =head2 backlog
 
@@ -343,8 +345,8 @@ L<Mojo::IOLoop> singleton.
   my $listen = $daemon->listen;
   $daemon    = $daemon->listen(['https://127.0.0.1:8080']);
 
-List of one or more locations to listen on, defaults to the value of the
-C<MOJO_LISTEN> environment variable or C<http://*:3000> (shortcut for
+Array reference with one or more locations to listen on, defaults to the value
+of the C<MOJO_LISTEN> environment variable or C<http://*:3000> (shortcut for
 C<http://0.0.0.0:3000>).
 
   # Listen on all IPv4 interfaces
@@ -359,15 +361,25 @@ C<http://0.0.0.0:3000>).
   # Listen on IPv4 and IPv6 interfaces
   $daemon->listen(['http://127.0.0.1:3000', 'http://[::1]:3000']);
 
+  # Listen on UNIX domain socket "/tmp/myapp.sock" (percent encoded slash)
+  $daemon->listen(['http+unix://%2Ftmp%2Fmyapp.sock']);
+
+  # File descriptor, as used by systemd
+  $daemon->listen(['http://127.0.0.1?fd=3']);
+
   # Allow multiple servers to use the same port (SO_REUSEPORT)
   $daemon->listen(['http://*:8080?reuse=1']);
 
   # Listen on two ports with HTTP and HTTPS at the same time
-  $daemon->listen([qw(http://*:3000 https://*:4000)]);
+  $daemon->listen(['http://*:3000', 'https://*:4000']);
 
   # Use a custom certificate and key
   $daemon->listen(['https://*:3000?cert=/x/server.crt&key=/y/server.key']);
 
+  # Domain specific certificates and keys (SNI)
+  $daemon->listen(
+    ['https://*:3000?example.com_cert=/x/my.crt&example.com_key=/y/my.key']);
+
   # Or even a custom certificate authority
   $daemon->listen(
     ['https://*:3000?cert=/x/server.crt&key=/y/server.key&ca=/z/ca.crt']);
@@ -380,11 +392,12 @@ These parameters are currently available:
 
   ca=/etc/tls/ca.crt
 
-Path to TLS certificate authority file.
+Path to TLS certificate authority file used to verify the peer certificate.
 
 =item cert
 
   cert=/etc/tls/server.crt
+  mojolicious.org_cert=/etc/tls/mojo.crt
 
 Path to the TLS cert file, defaults to a built-in test certificate.
 
@@ -392,11 +405,19 @@ Path to the TLS cert file, defaults to a built-in test certificate.
 
   ciphers=AES128-GCM-SHA256:RC4:HIGH:!MD5:!aNULL:!EDH
 
-Cipher specification string.
+TLS cipher specification string. For more information about the format see
+L<https://www.openssl.org/docs/manmaster/apps/ciphers.html#CIPHER-STRINGS>.
+
+=item fd
+
+  fd=3
+
+File descriptor with an already prepared listen socket.
 
 =item key
 
   key=/etc/tls/server.key
+  mojolicious.org_key=/etc/tls/mojo.key
 
 Path to the TLS key file, defaults to a built-in test key.
 
@@ -407,29 +428,42 @@ Path to the TLS key file, defaults to a built-in test key.
 Allow multiple servers to use the same port with the C<SO_REUSEPORT> socket
 option.
 
+=item single_accept
+
+  single_accept=1
+
+Only accept one connection at a time.
+
 =item verify
 
   verify=0x00
 
-TLS verification mode, defaults to C<0x03>.
+TLS verification mode, defaults to C<0x03> if a certificate authority file has
+been provided, or C<0x00>.
+
+=item version
+
+  version=TLSv1_2
+
+TLS protocol version.
 
 =back
 
 =head2 max_clients
 
   my $max = $daemon->max_clients;
-  $daemon = $daemon->max_clients(1000);
+  $daemon = $daemon->max_clients(100);
 
-Maximum number of concurrent connections this server is allowed to handle
-before stopping to accept new incoming connections, passed along to
-L<Mojo::IOLoop/"max_connections">.
+Maximum number of accepted connections this server is allowed to handle
+concurrently, before stopping to accept new incoming connections, passed along
+to L<Mojo::IOLoop/"max_connections">.
 
 =head2 max_requests
 
   my $max = $daemon->max_requests;
-  $daemon = $daemon->max_requests(100);
+  $daemon = $daemon->max_requests(250);
 
-Maximum number of keep-alive requests per connection, defaults to C<25>.
+Maximum number of keep-alive requests per connection, defaults to C<100>.
 
 =head2 silent
 
@@ -443,27 +477,40 @@ Disable console messages.
 L<Mojo::Server::Daemon> inherits all methods from L<Mojo::Server> and
 implements the following new ones.
 
+=head2 ports
+
+  my $ports = $daemon->ports;
+
+Get all ports this server is currently listening on.
+
+  # All ports
+  say for @{$daemon->ports};
+
 =head2 run
 
   $daemon->run;
 
-Run server.
+Run server and wait for L</"SIGNALS">.
 
 =head2 start
 
   $daemon = $daemon->start;
 
-Start accepting connections.
+Start or resume accepting connections through L</"ioloop">.
 
   # Listen on random port
-  my $id   = $daemon->listen(['http://127.0.0.1'])->start->acceptors->[0];
-  my $port = $daemon->ioloop->acceptor($id)->port;
+  my $port = $daemon->listen(['http://127.0.0.1'])->start->ports->[0];
+
+  # Run multiple web servers concurrently
+  my $daemon1 = Mojo::Server::Daemon->new(listen => ['http://*:3000'])->start;
+  my $daemon2 = Mojo::Server::Daemon->new(listen => ['http://*:4000'])->start;
+  Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
 
 =head2 stop
 
   $daemon = $daemon->stop;
 
-Stop accepting connections.
+Stop accepting connections through L</"ioloop">.
 
 =head1 DEBUGGING
 
@@ -474,6 +521,6 @@ diagnostics information printed to C<STDERR>.
 
 =head1 SEE ALSO
 
-L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicio.us>.
+L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicious.org>.
 
 =cut
